# Generated by roxygen2: do not edit by hand

export(dmhn)
export(mhn_kurtosis)
export(mhn_mean)
export(mhn_mode)
export(mhn_skewness)
export(mhn_var)
export(pmhn)
export(qmhn)
export(rmhn)
importFrom(Rcpp,sourceCpp)
importFrom(stats,dgamma)
importFrom(stats,dnorm)
importFrom(stats,integrate)
importFrom(stats,pgamma)
importFrom(stats,pnorm)
importFrom(stats,qgamma)
importFrom(stats,qnorm)
importFrom(stats,rgamma)
importFrom(stats,rnorm)
importFrom(stats,uniroot)
useDynLib(mhn, .registration = TRUE)
